/*
 * crash-manager
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <assert.h>
#include <iniparser.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "defs.h"
#include "log.h"
#include "util.h"

static const char *const report_type_strmap[] = {
	[REP_TYPE_INFO] = "INFO",
	[REP_TYPE_FULL] = "FULL",
};

static const char *report_type_to_str(enum ReportType report_type)
{
	if (report_type < 0 || report_type >= ARRAY_SIZE(report_type_strmap))
		return NULL;
	return report_type_strmap[report_type];
}

enum ReportType report_type_from_str(const char *report_type_str)
{
	assert(report_type_str);

	for (int i = 0; i < (int)ARRAY_SIZE(report_type_strmap); i++) {
		if (0 == strcmp(report_type_str, report_type_strmap[i]))
			return (enum ReportType)i;
	}

	return REP_TYPE_INVALID;
}

bool config_init(config_t *c, const char *const path)
{
	assert(c);
	assert(path);

	dictionary *ini = iniparser_load(path);
	if (!ini) {
		_E("Failed to load config file %s", path);
		return false;
	}

	bool ret = false;

#define GET(type, key, defval) iniparser_get##type(ini, CRASH_SECTION ":" key, defval)

	c->crash_root_path = strdup(GET(string, "CrashRootPath", CRASH_ROOT_PATH));
	if (!c->crash_root_path)
		goto out;

	/* strdup() can technically fail, but we don't mind. It is better to
	 * create a report without the extra script than to abort completely. */
	char *extrascript = GET(string, "ExtraScript", NULL);
	c->extra_script = extrascript ? strdup(extrascript) : NULL;

	char *reptype = GET(string, "ReportType", (char *)report_type_strmap[REP_TYPE_FULL]);
	c->report_type = report_type_from_str(reptype);
	if (!report_type_to_str(c->report_type))
		goto out;

	c->system_max_use = GET(int, "SystemMaxUse", SYSTEM_MAX_USE);
	c->system_keep_free = GET(int, "SystemKeepFree", SYSTEM_KEEP_FREE);
	c->max_retention_sec = GET(int, "MaxRetentionSec", MAX_RETENTION_SEC);
	c->max_crash_dump = GET(int, "MaxCrashDump", MAX_CRASH_DUMP);
	c->dump_core = GET(boolean, "DumpCore", DUMP_CORE);
	c->allow_zip = GET(boolean, "AllowZip", ALLOW_ZIP);

#undef GET

	ret = true;
out:
	iniparser_freedict(ini);
	return ret;
}

void config_free(config_t *c)
{
	assert(c);

	free(c->crash_root_path);
	free(c->extra_script);
}
